﻿#ifndef __BOUNDINGBOX_H__
#define __BOUNDINGBOX_H__

#include "Vector2D.h"
#include "Vector3D.h"
#include <cfloat>

class BoundingBox
{
public:
	enum BoxPoints
	{
		LEFT_DOWN_FRONT,
		LEFT_UP_FRONT,
		RIGHT_UP_FRONT,
		RIGHT_DOWN_FRONT,
		LEFT_DOWN_BACK,
		LEFT_UP_BACK,
		RIGHT_UP_BACK,
		RIGHT_DOWN_BACK,
		TOTAL_POINTS
	};
	Vector3D minPoint;
	Vector3D maxPoint;
	INLINE BoundingBox(const Vector3D& minP, const Vector3D& maxP)
	{
		minPoint = minP;
		maxPoint = maxP;
	}
	INLINE BoundingBox()
	{
		Reset();
	}
	void Reset()
	{
		maxPoint = Vector3D(- FLT_MAX, - FLT_MAX, - FLT_MAX);
		minPoint = Vector3D(FLT_MAX, FLT_MAX, FLT_MAX);	
	}
	void CalcPoints()
	{
	}
	template<typename T>
	INLINE void Translate(T x, T y, T z)
	{
//		assert(IsValid() && "Invalid Bounding Box");
		minPoint.x += static_cast<float>(x);
		minPoint.y += static_cast<float>(y);
		minPoint.z += static_cast<float>(z);
		maxPoint.x += static_cast<float>(x);
		maxPoint.y += static_cast<float>(y);
		maxPoint.z += static_cast<float>(z);
	}
	INLINE void Grow(float x, float y, float z)
	{
		minPoint.x -= x;
		minPoint.y -= y;
		minPoint.z -= z;
		maxPoint.x += x;
		maxPoint.y += y;
		maxPoint.z += z;
	}
	INLINE void Grow(const Vector3D& offset)
	{
		Grow(offset.x, offset.y, offset.z);
	}
	INLINE void Translate(const Vector3D& pos)
	{
		Translate(pos.x, pos.y, pos.z);
	}
	INLINE void Scale(float scaleX, float scaleY, float scaleZ)
	{
//		assert(scaleX > 0.0f && "Invalid value");
		//assert(scaleY > 0.0f && "Invalid value");
	///	assert(scaleZ > 0.0f && "Invalid value");
	//	assert(IsValid() && "Invalid Bounding Box");

		Vector3D center = GetCenter();
		Translate(-center);
	//	assert(IsValid() && "Invalid Bounding Box");
		float minX = minPoint.x * scaleX;
		float minY = minPoint.y * scaleY;
		float minZ = minPoint.z * scaleZ;
		float maxX = maxPoint.x * scaleX;
		float maxY = maxPoint.y * scaleY;
		float maxZ = maxPoint.z * scaleZ;
		Reset();
		AddVertex(minX, minY, minZ);
		AddVertex(maxX, maxY, maxZ);
		//assert(IsValid() && "Invalid Bounding Box");
		Translate(center);
		//assert(IsValid() && "Invalid Bounding Box");
	}
	INLINE void Scale(const Vector3D& scale)
	{
		Scale(scale.x, scale.y, scale.z);
	}
	INLINE void Scale(float value)
	{
		Scale(value, value, value);
	}
	INLINE bool IsValid() const
	{
		return maxPoint.x >= minPoint.x && maxPoint.y >= minPoint.y && maxPoint.z >= minPoint.z;
	}
	INLINE bool IsZero() const
	{
		return minPoint == maxPoint;
	}
	// 
	INLINE bool Contains(const BoundingBox& aabb) const
	{
		const Vector3D& maxP = aabb.maxPoint;
		const Vector3D& minP = aabb.minPoint;
		return maxPoint.x >= maxP.x && maxPoint.y >= maxP.y && maxPoint.z >= maxP.z && minPoint.x <= minP.x && minPoint.y <= minP.y && minPoint.z <= minP.z;
	}
	//
	Vector3D GetPoint(uint pointNumber) const
	{
		switch (pointNumber)
		{
			case LEFT_DOWN_FRONT:
			{
				// left bottom near
				return minPoint;
			}
			break;
			case LEFT_UP_FRONT:
			{
				// left top near
				return Vector3D(minPoint.x, maxPoint.y, minPoint.z);
			}
			break;
			case RIGHT_UP_FRONT:
			{
				// right top near
				return Vector3D(maxPoint.x, maxPoint.y, minPoint.z);
			}
			break;
			case RIGHT_DOWN_FRONT:
			{
				// right bottom near
				return Vector3D(maxPoint.x, minPoint.y, minPoint.z);
			}
			break;
			case LEFT_DOWN_BACK:
			{
				// left bottom far
				return Vector3D(minPoint.x, minPoint.y, maxPoint.z);
			}
			break;
			case LEFT_UP_BACK:
			{
				// left top far
				return Vector3D(minPoint.x, maxPoint.y, maxPoint.z);
			}
			break;
			case RIGHT_UP_BACK:
			{
				// right top far
				return maxPoint;
			}
			break;
			case RIGHT_DOWN_BACK:
			{
				// right bottom far
				return Vector3D(maxPoint.x, minPoint.y, maxPoint.z);
			}
			break;
		};
		//assert(!"Invalid point number");
		return Vector3D::Zero();
	}
	void GetPoints(Vector3D* outPoints) const
	{
	//	assert(IsValid() && "Invalid Bounding Box");
		const Vector3D points[TOTAL_POINTS] = 
		{	minPoint,
			Vector3D(minPoint.x, maxPoint.y, minPoint.z),
			Vector3D(maxPoint.x, maxPoint.y, minPoint.z),
			Vector3D(maxPoint.x, minPoint.y, minPoint.z),
			Vector3D(minPoint.x, minPoint.y, maxPoint.z),
			Vector3D(minPoint.x, maxPoint.y, maxPoint.z),
			maxPoint,
			Vector3D(maxPoint.x, minPoint.y, maxPoint.z)
		};
	//	assert(outPoints && "NULL Pointer");
		memcpy(outPoints, points, sizeof(points));
	}
	float Get2DRadius() const
	{
		Vector2D r = Vector2D(maxPoint.x - minPoint.x, maxPoint.z - minPoint.z) * 0.5f;
		float l = r.Length();
		return l;
	}
	float GetRadius() const
	{
		Vector3D r = (maxPoint - minPoint) * 0.5f;
		float l = r.Length();
		return l;
	}
	// checking containing a point in a bounding box
	INLINE bool Contains(float x, float z) const
	{
		// checking range
		if (x < minPoint.x)
		{
			return false;
		}
		if (z < minPoint.z)
		{
			return false;
		}
		if (x > maxPoint.x)
		{
			return false;
		}
		if (z > maxPoint.z)
		{
			return false;
		}
		return true;
	}
	// checking of collision of two bounding box
	INLINE bool Contains(float x, float y, float z) const
	{
		// checking range
		if (x < minPoint.x) 
		{
			return false;
		}
		if (y < minPoint.y) 
		{
			return false;
		}
		if (z < minPoint.z) 
		{
			return false;
		}
		if (x > maxPoint.x)
		{
			return false;
		}
		if (y > maxPoint.y)
		{
			return false;
		}
		if (z > maxPoint.z)
		{
			return false;
		}
		return true;
	}
	// checking of collision of two bounding box
	template<typename Box>
	INLINE bool Intersect(const Box& aabb) const
	{
		if (minPoint.x > maxPoint.x)
		{
			return false;
		}
		if (maxPoint.x < aabb.minPoint.x)
		{
			return false;
		}
		if (minPoint.y > aabb.maxPoint.y)
		{
			return false;
		}
		if (maxPoint.y < aabb.minPoint.y)
		{
			return false;
		}
		if (minPoint.z > aabb.maxPoint.z)
		{
			return false;
		}
		if (maxPoint.z < aabb.minPoint.z)
		{
			return false;
		}
		return true;
	}
	// checking containing a point in a bounding box
	INLINE bool Contains(const Vector3D& p) const
	{
		return Contains(p.x, p.y, p.z);
	}
	INLINE bool Contains(const Vector2D& p) const
	{
		return Contains(p.x, minPoint.y, p.y);
	}
	//
	INLINE const Vector3D& GetMinPoint() const
	{
		return minPoint;
	}
	//
	INLINE const Vector3D& GetMaxPoint() const
	{
		return maxPoint;
	}
	template<typename BoxType>
	void INLINE Add(const BoxType& box)
	{
	//	assert(box.IsValid() && "Invalid Bounding Box");
		AddVertex(box.GetMinPoint());
		AddVertex(box.GetMaxPoint());
	}
	// adding vertex to computing a bounding box
	template<typename T>
	INLINE void AddVertex(T x, T y, T z)
	{
		AddVertex(static_cast<float>(x), static_cast<float>(y), static_cast<float>(z));
	}
	// adding vertex to computing a bounding box
	INLINE void AddVertex(float x, float y, float z)
	{
		// checking range
		if (minPoint.x > x)
		{
			minPoint.x = x;
		}
		if (minPoint.y > y)
		{
			minPoint.y = y;
		}
		if (minPoint.z > z)
		{
			minPoint.z = z;
		}

		if (maxPoint.x < x)
		{
			maxPoint.x = x;
		}
		if (maxPoint.y < y)
		{
			maxPoint.y = y;
		}
		if (maxPoint.z < z)
		{
			maxPoint.z = z;
		}
	}
	// adding vertex to computing a bounding box
	void AddVertex(const float* vertex)
	{
		//assert(vertex && "NULL Pointer");
		AddVertex(vertex[0], vertex[1], vertex[2]);
	}
	//
	void AddVertex(const Vector3D& vert)
	{
		AddVertex(vert.x, vert.y, vert.z);
	}
	//
	void AddLineSegmentVertex(const Vector3D* segment)
	{
	//	assert(segment && "NULL Pointer");
		AddVertex(segment[0]);
		AddVertex(segment[1]);
	}
	//
	void Merge(const BoundingBox& box)
	{
		Add(box);
	}
	//
	INLINE void SetMinPoint(float x, float y, float z)
	{
		minPoint.x = x;
		minPoint.y = y;
		minPoint.z = z;
	}
	//
	INLINE void SetMaxPoint(float x, float y, float z)
	{
		maxPoint.x = x;
		maxPoint.y = y;
		maxPoint.z = z;
	}
	//
	INLINE void SetMinPoint(const Vector3D& newPoint)
	{
		minPoint = newPoint;
	}
	//
	INLINE void SetMaxPoint(const Vector3D& newPoint)
	{
		maxPoint = newPoint;
	}
	//
	INLINE const Vector3D GetCenter() const
	{
		return (maxPoint + minPoint) * 0.5f;
	}
	//
	INLINE Vector3D GetRadiusVector() const
	{
		return maxPoint - GetCenter();
	}
	//
	INLINE float GetRadiusLen() const
	{
		Vector3D v = GetRadiusVector();
		return v.Length();
	}
	//
	INLINE const Vector3D GetSize() const
	{
		return maxPoint - minPoint;
	}
	//
	INLINE void InitBoundingBox(const Vector3D& MinPoint, const Vector3D& MaxPoint)
	{
	//	assert(MinPoint.x <= MaxPoint.x && "Invalid Value");
	//	assert(MinPoint.y <= MaxPoint.y && "Invalid Value");
	//	assert(MinPoint.z <= MaxPoint.z && "Invalid Value");
		minPoint = MinPoint;
		maxPoint = MaxPoint;
	}
	//
	Vector3D GetNearestPointFromPosition(const Vector3D& pos) const
	{
		Vector3D points[TOTAL_POINTS];
		GetPoints(points);
		float minDist = FLT_MAX;
		int index = 0;
		for(int i = 0; i < TOTAL_POINTS; ++i)
		{
			float dist = pos.GetDistanceSqFrom(points[i]);
			if (dist < minDist)
			{
				minDist = dist;
				index = i;
			}
		}
	//	assert(index < TOTAL_POINTS && "Out Of Range");
		return points[index];
	}
	//
	Vector3D GetFarPointFromPosition(const Vector3D& pos) const
	{
		Vector3D points[TOTAL_POINTS];
		GetPoints(points);
		float maxDist = 0;
		int index = 0;
		for(int i = 0; i < TOTAL_POINTS; ++i)
		{
			float dist = pos.GetDistanceSqFrom(points[i]);
			if (dist > maxDist)
			{
				maxDist = dist;
				index = i;
			}
		}
	//	assert(index < TOTAL_POINTS && "Out Of Range");
		return points[index];
	}
	//
	float GetNearestDistFromPosition(const Vector3D& pos) const
	{
		const Vector3D Point = GetNearestPointFromPosition(pos);
		return Point.GetDistanceFrom(pos);
	}
	//
	float GetFarDistFromPosition(const Vector3D& pos) const
	{
		const Vector3D Point = GetFarPointFromPosition(pos);
		return Point.GetDistanceFrom(pos);
	}
	//
	float GetDistanceToCenter(const Vector3D& pos) const
	{
		const Vector3D Point = GetCenter();
		return Point.GetDistanceFrom(pos);
	}
	//
	template<typename T>
	bool operator == (const T& box) const
	{
		return maxPoint == box.maxPoint && minPoint == box.minPoint;
	}
	//
	template<typename T>
	bool operator != (const T& box) const
	{
		return maxPoint != box.maxPoint || minPoint != box.minPoint;
	}
	//
	template<typename T>
	BoundingBox& operator = (const T& box)
	{
		minPoint = box.minPoint;
		maxPoint = box.maxPoint;
		return *this;
	}
};

#endif
