/****************************************************************************
**
** Copyright (C) 2016 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the examples of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:BSD$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** BSD License Usage
** Alternatively, you may use this file under the terms of the BSD license
** as follows:
**
** "Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are
** met:
**   * Redistributions of source code must retain the above copyright
**     notice, this list of conditions and the following disclaimer.
**   * Redistributions in binary form must reproduce the above copyright
**     notice, this list of conditions and the following disclaimer in
**     the documentation and/or other materials provided with the
**     distribution.
**   * Neither the name of The Qt Company Ltd nor the names of its
**     contributors may be used to endorse or promote products derived
**     from this software without specific prior written permission.
**
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
** "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
** LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
** A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
** OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
** SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
** LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
**
** $QT_END_LICENSE$
**
****************************************************************************/

#include "glwidget.h"
#include <QMouseEvent>
#include <QOpenGLShaderProgram>
#include <QCoreApplication>
#include <math.h>
#include <android/log.h>
#include "VertexFormat.h"
#include "GeometryData.h"

#define LOG(x)

VertexData gridData;

const char* tag = "ARM_OGL_ES2";

int CheckOpenGLErrorImpl(const char* file = 0, uint line = 0)
{
	//
	int err = glGetError();
	//
	if (GL_NO_ERROR != err)
	{
		struct OutMessage
		{
			OutMessage(const char* error, const char* file, uint line)
			{
				if (file && line)
				{
					__android_log_print(ANDROID_LOG_DEBUG, tag, "%s, file %s, line %d", error, file, line);
				}
				else
				{
					__android_log_print(ANDROID_LOG_DEBUG, tag, error);
				}
			}
		};
		switch(err)
		{
			case GL_INVALID_OPERATION:
			{
				OutMessage("3DEngine OpenGL Error: invalid operation", file, line);
			}
			break;
			case GL_INVALID_VALUE:
			{
				OutMessage("3DEngine OpenGL Error: invalid value", file, line);
			}
			break;
			case GL_INVALID_ENUM:
			{
				OutMessage("3DEngine OpenGL Error: invalid enum", file, line);
			}
			break;
			case GL_INVALID_FRAMEBUFFER_OPERATION:
			{
				OutMessage("3DEngine OpenGL Error: invalid frame buffer operation", file, line);
			}
			break;
			case GL_OUT_OF_MEMORY:
			{
				OutMessage("3DEngine OpenGL Error: out of memory", file, line);
			}
			break;
		}
	}
	return err;
}

void OutOpenGLInfo()
{
	const char* str = reinterpret_cast<const char *>(glGetString(GL_RENDERER));
	if (str)
	{
		// renderer name
		__android_log_print(ANDROID_LOG_DEBUG, tag,"OpenGL Renderer: %s", str);
		CheckOpenGLErrorImpl();
	}
	str = reinterpret_cast<const char *>(glGetString(GL_VENDOR));
	if (str)
	{
		 // vendor name
		__android_log_print(ANDROID_LOG_DEBUG, tag, "OpenGL Vendor: %s", str);
		CheckOpenGLErrorImpl();
	}
	str = reinterpret_cast<const char *>(glGetString(GL_VERSION));
	if (str)
	{
		// API version
		__android_log_print(ANDROID_LOG_DEBUG, tag, "OpenGL Version: %s", str);
		CheckOpenGLErrorImpl();
	}
	const char* ext = reinterpret_cast<const char *>(glGetString(GL_EXTENSIONS));
	//
	__android_log_print(0, tag, "OpenGL extensions: %s", ext);
}

void DrawLinesImpl(const VertexData& gridData)
{
	glBindBuffer(GL_ARRAY_BUFFER, gridData.vb);
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(GLfloat), 0);
	glDrawArrays(GL_LINES, 0, gridData.numVertices);
}

GLWidget::GLWidget(QWidget *parent)
    : QOpenGLWidget(parent),
      m_xRot(0),
      m_yRot(0),
      m_zRot(0),
      m_program(0)
{
	QSurfaceFormat fmt = format();
	fmt.setAlphaBufferSize(8);
	setFormat(fmt);
}

GLWidget::~GLWidget()
{
    cleanup();
}

QSize GLWidget::minimumSizeHint() const
{
    return QSize(50, 50);
}

QSize GLWidget::sizeHint() const
{
    return QSize(400, 400);
}

static void qNormalizeAngle(int &angle)
{
    while (angle < 0)
        angle += 360 * 16;
    while (angle > 360 * 16)
        angle -= 360 * 16;
}

void GLWidget::setXRotation(int angle)
{
    qNormalizeAngle(angle);
    if (angle != m_xRot) {
        m_xRot = angle;
        emit xRotationChanged(angle);
        update();
    }
}

void GLWidget::setYRotation(int angle)
{
    qNormalizeAngle(angle);
    if (angle != m_yRot) {
        m_yRot = angle;
        emit yRotationChanged(angle);
        update();
    }
}

void GLWidget::setZRotation(int angle)
{
    qNormalizeAngle(angle);
    if (angle != m_zRot) {
        m_zRot = angle;
        emit zRotationChanged(angle);
        update();
    }
}

void GLWidget::cleanup()
{
    makeCurrent();
    m_logoVbo.destroy();
    delete m_program;
    m_program = 0;
    doneCurrent();
}

static const char *vertexShaderSource =
    "attribute vec4 vertex;\n"
    "uniform mat4 projMatrix;\n"
    "uniform mat4 mvMatrix;\n"
	//"uniform mat3 normalMatrix;\n"
    "void main() {\n"
    "   gl_Position = projMatrix * mvMatrix * vertex;\n"
    "}\n";

static const char *fragmentShaderSource =
    "varying highp vec3 vert;\n"
    "void main() {\n"
		"   gl_FragColor = vec4(0.0, 1.0, 0.0, 1.0);\n"
    "}\n";

void GLWidget::initializeGL()
{
    // In this example the widget's corresponding top-level window can change
    // several times during the widget's lifetime. Whenever this happens, the
    // QOpenGLWidget's associated context is destroyed and a new one is created.
    // Therefore we have to be prepared to clean up the resources on the
    // aboutToBeDestroyed() signal, instead of the destructor. The emission of
    // the signal will be followed by an invocation of initializeGL() where we
    // can recreate all resources.
    connect(context(), &QOpenGLContext::aboutToBeDestroyed, this, &GLWidget::cleanup);

    initializeOpenGLFunctions();
	OutOpenGLInfo();
	glClearColor(0, 0, 0, 1);
	CheckOpenGLErrorImpl();

    m_program = new QOpenGLShaderProgram;
	m_program->addShaderFromSourceCode(QOpenGLShader::Vertex, vertexShaderSource);
	m_program->addShaderFromSourceCode(QOpenGLShader::Fragment, fragmentShaderSource);
    m_program->bindAttributeLocation("vertex", 0);
    m_program->link();

    m_program->bind();
    m_projMatrixLoc = m_program->uniformLocation("projMatrix");
    m_mvMatrixLoc = m_program->uniformLocation("mvMatrix");

    // Create a vertex array object. In OpenGL ES 2.0 and OpenGL 2.x
    // implementations this is optional and support may not be present
    // at all. Nonetheless the below code works in all cases and makes
    // sure there is a VAO when one is needed.
	//m_vao.create();
	//QOpenGLVertexArrayObject::Binder vaoBinder(&m_vao);

    // Setup our vertex buffer object.
    m_logoVbo.create();
	CheckOpenGLErrorImpl();
    m_logoVbo.bind();
	CheckOpenGLErrorImpl();
    m_logoVbo.allocate(m_logo.constData(), m_logo.count() * sizeof(GLfloat));
	CheckOpenGLErrorImpl();

    // Store the vertex attribute bindings for the program.
    setupVertexAttribs();

    // Our camera never changes in this example.
    m_camera.setToIdentity();
    m_camera.translate(0, 0, -1);

    m_program->release();

	const float maxX = 100;
	const float maxZ = 100;
	InitLinesGrid(maxX, maxZ, 0.0f, 1.0);
}

void GLWidget::setupVertexAttribs()
{
    m_logoVbo.bind();
    QOpenGLFunctions *f = QOpenGLContext::currentContext()->functions();
    f->glEnableVertexAttribArray(0);
	CheckOpenGLErrorImpl();
    f->glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(GLfloat), 0);
	CheckOpenGLErrorImpl();
}

void GLWidget::CreateBuffer(const void* data, size_t size, uint& vb)
{
	//
	glGenBuffers(1, &vb);
	CheckOpenGLErrorImpl();
	glBindBuffer(GL_ARRAY_BUFFER, vb);
	CheckOpenGLErrorImpl();
	glBufferData(GL_ARRAY_BUFFER, size, data, GL_STATIC_DRAW);
	CheckOpenGLErrorImpl();
}

void GLWidget::InitLinesGrid(float MaxCoordX, float MaxCoordZ, float y, float scale)
{
	if (!MaxCoordX && !MaxCoordZ)
	{
		const float Val = 100.0f;
		MaxCoordX = Val;
		MaxCoordZ = Val;
	}
	if (!MaxCoordX)
	{
		MaxCoordX = MaxCoordZ;
	}
	if (!MaxCoordZ)
	{
		MaxCoordZ = MaxCoordX;
	}
	float MinCoord = std::min(MaxCoordX, MaxCoordZ);
	float k = 0.0f;
	if (scale)
	{
		//k = scale / 5.0f;
		k = scale;
	}
	else
	{
		const float minK = 10.0f;
		k = minK;
	}
	float Step = ceilf(MinCoord / k);
	float MaxCoord = std::max(MaxCoordX, MaxCoordZ);
	MaxCoord = ceilf(MaxCoord / Step) * Step;

	float startX = -MaxCoord;
	float endX = MaxCoord;
	float startZ = -MaxCoord;
	float endZ = MaxCoord;

	struct Line3D
	{
		Vector3D start;
		Vector3D end;
	};
	size_t gridSize = static_cast<size_t>(4 * (MaxCoord * MaxCoord) / Step);
	static const size_t maxValue = 15000000;
	if (gridSize > maxValue) // to do
	{
		return;
	}
	std::vector<Line3D> LineGrids;
	LineGrids.resize(gridSize + 4);
	//if (!LineGrids.begin())
	//{
	//	LOG("3DEngine: Create Grid: Out Of Memory");
	//	return;
	//}
	VertexPrimitiveType i = 0;
#if 1
	static float coord = 100.0f;
	startX = coord;
	endX = -startX;
	startZ = -coord;
	endZ = -startZ;
	static float step = 10.0f;
	Step = step;
#endif
	// X
	for(float x = startX; x <= endX; x += Step)
	{
		Line3D& line = LineGrids[i++];

		line.start = Vector3D(x, y, startZ);
		line.end = Vector3D(x, y, endZ);
	}
	// Z
	for(float z = startZ; z <= endZ; z += Step)
	{
		Line3D& line = LineGrids[i++];
		line.start = Vector3D(startX, y, z);
		line.end = Vector3D(endX, y, z);
	}
	gridData.numPrimitives = i;
	gridData.numVertices = gridData.numPrimitives << 1;

	//gridData.y = y;
//	gridData.scale = scale;
//	gridData.MaxCoordX = MaxCoordX;
//	gridData.MaxCoordZ = MaxCoordZ;
	//
	CreateBuffer(&LineGrids[0], gridSize * sizeof(Line3D), gridData.vb);
}

void GLWidget::DrawLines(const VertexData& gridData)
{
#if 1
	glBindBuffer(GL_ARRAY_BUFFER, gridData.vb);
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(GLfloat), 0);
	glDrawArrays(GL_LINES, 0, gridData.numVertices);
#else
	DrawLinesImpl(gridData);
#endif
}

void GLWidget::paintGL()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	CheckOpenGLErrorImpl();
    glEnable(GL_DEPTH_TEST);
	CheckOpenGLErrorImpl();
    glEnable(GL_CULL_FACE);
	CheckOpenGLErrorImpl();

    m_world.setToIdentity();
    m_world.rotate(180.0f - (m_xRot / 16.0f), 1, 0, 0);
    m_world.rotate(m_yRot / 16.0f, 0, 1, 0);
    m_world.rotate(m_zRot / 16.0f, 0, 0, 1);

    m_program->bind();
    m_program->setUniformValue(m_projMatrixLoc, m_proj);
    m_program->setUniformValue(m_mvMatrixLoc, m_camera * m_world);

	DrawLines(gridData);

	m_logoVbo.bind();
	CheckOpenGLErrorImpl();
	setupVertexAttribs();
	glDrawArrays(GL_TRIANGLES, 0, m_logo.vertexCount());
	CheckOpenGLErrorImpl();
	m_logoVbo.release();
	CheckOpenGLErrorImpl();

    m_program->release();
}

void GLWidget::resizeGL(int w, int h)
{
    m_proj.setToIdentity();
    m_proj.perspective(45.0f, GLfloat(w) / h, 0.01f, 100.0f);
}

void GLWidget::mousePressEvent(QMouseEvent *event)
{
    m_lastPos = event->pos();
}

void GLWidget::mouseMoveEvent(QMouseEvent *event)
{
    int dx = event->x() - m_lastPos.x();
    int dy = event->y() - m_lastPos.y();

    if (event->buttons() & Qt::LeftButton) {
        setXRotation(m_xRot + 8 * dy);
        setYRotation(m_yRot + 8 * dx);
    } else if (event->buttons() & Qt::RightButton) {
        setXRotation(m_xRot + 8 * dy);
        setZRotation(m_zRot + 8 * dx);
    }
    m_lastPos = event->pos();
}
